//
// AXI bus multiplexor
// Input: three slave ports
// Output: one master port
// Round robin port selection.
//
// 
// Design by G.J. van Loo, FenLogic Ltd, October-2017.
// Mail: verilog@fenlogic.com
//
// This program is free software. It comes without any guarantees or
// warranty to the extent permitted by applicable law. Although the
// author has attempted to find and correct any bugs in this free software
// program, the author is not responsible for any damage or losses of any
// kind caused by the use or misuse of the program. You can redistribute
// the program and or modify it in any form without obligations, but the
// author would appreciated if these credits stays in.
//
// This mux is about as small as I can make it.
// It is mostly combinatorial (Only 6 registers)
// To keep logic minimal two extra ID bits are added to the output. 
//
// There is no user port but it can be added alongside
// the signals like cache, lock, prot which are all
// handled in an identical way
//
// Efficiency: 
// When more then one slave is requesting service this mux will be 
// operative 100% of the time. There are no wasted clock cycles.
// A single clck cycle is wasted whan the mux is idle and has
// to switch to a different imput port. 

module axi_mux3rr
#(parameter ADRS_BITS = 32,
            DATA_BITS = 32,
            ID_BITS   =  6, // Slave port ID Bits 
            LEN_BITS  =  4  // Len bits 
 )
(
   input              clk    ,
   input              reset_n,
   
   
   /******************\
    *                *
    *  SLAVE PORT 0  *
    *                *
   \******************/
   
   // write address 
   input                    s0_awvalid , 
   output reg               s0_awready , 
   input    [ADRS_BITS-1:0] s0_awaddr  , 
   input     [LEN_BITS-1:0] s0_awlen   , 
   input             [ 2:0] s0_awsize  , 
   input      [ID_BITS-1:0] s0_awid    , 
   input             [ 1:0] s0_awburst , 
   input             [ 3:0] s0_awcache , 
   input             [ 1:0] s0_awlock  , 
   input             [ 2:0] s0_awprot  , 
   input             [ 3:0] s0_awqos   , 
                           
   // write data           
   input                    s0_wvalid  , 
   output reg               s0_wready  , 
   input    [DATA_BITS-1:0] s0_wdata   , 
   input             [ 3:0] s0_wstrb   , 
   input                    s0_wlast   , 
   input      [ID_BITS-1:0] s0_wid     , 

   // write response    
   output                   s0_bvalid  , 
   input                    s0_bready  , 
   output           [ 1 :0] s0_bresp   , 
   output     [ID_BITS-1:0] s0_bid     , 


   //
   // read interface
   //
   
   // read address 
   input                    s0_arvalid, 
   output reg               s0_arready, 
   input   [ADRS_BITS-1: 0] s0_araddr , 
   input      [ID_BITS-1:0] s0_arid   , 
   input     [LEN_BITS-1:0] s0_arlen  , 
   input            [ 2: 0] s0_arsize , 
   input            [ 1: 0] s0_arburst, 
   input            [ 3: 0] s0_arcache, 
   input            [ 1: 0] s0_arlock , 
   input            [ 2: 0] s0_arprot , 
   input            [ 3: 0] s0_arqos  , 

   // read data
   output                   s0_rvalid , 
   input                    s0_rready , 
   output   [DATA_BITS-1:0] s0_rdata  , 
   output                   s0_rlast  , 
   output     [ID_BITS-1:0] s0_rid    , 
   output           [ 1 :0] s0_rresp  , 

  /******************\
   *                *
   *  SLAVE PORT 1  *
   *                *
  \******************/
   
  // write address 
   input                    s1_awvalid , 
   output reg               s1_awready , 
   input    [ADRS_BITS-1:0] s1_awaddr  , 
   input     [LEN_BITS-1:0] s1_awlen   , 
   input             [ 2:0] s1_awsize  , 
   input      [ID_BITS-1:0] s1_awid    , 
   input             [ 1:0] s1_awburst , 
   input             [ 3:0] s1_awcache , 
   input             [ 1:0] s1_awlock  , 
   input             [ 2:0] s1_awprot  , 
   input             [ 3:0] s1_awqos   , 
                           
   // write data           
   input                    s1_wvalid  , 
   output reg               s1_wready  , 
   input    [DATA_BITS-1:0] s1_wdata   , 
   input             [ 3:0] s1_wstrb   , 
   input                    s1_wlast   , 
   input      [ID_BITS-1:0] s1_wid     , 

   // write response    
   output                   s1_bvalid  , 
   input                    s1_bready  , 
   output           [ 1 :0] s1_bresp   , 
   output     [ID_BITS-1:0] s1_bid     , 


   //
   // read interface
   //
   
   // read address 
   input                    s1_arvalid, 
   output reg               s1_arready, 
   input   [ADRS_BITS-1: 0] s1_araddr , 
   input      [ID_BITS-1:0] s1_arid   , 
   input     [LEN_BITS-1:0] s1_arlen  , 
   input            [ 2: 0] s1_arsize , 
   input            [ 1: 0] s1_arburst, 
   input            [ 3: 0] s1_arcache, 
   input            [ 1: 0] s1_arlock , 
   input            [ 2: 0] s1_arprot , 
   input            [ 3: 0] s1_arqos  , 

   // read data
   output                   s1_rvalid , 
   input                    s1_rready , 
   output   [DATA_BITS-1:0] s1_rdata  , 
   output                   s1_rlast  , 
   output     [ID_BITS-1:0] s1_rid    , 
   output           [ 1 :0] s1_rresp  , 
 
  /******************\
   *                *
   *  SLAVE PORT 2  *
   *                *
  \******************/
   
   // write address 
   input                    s2_awvalid , 
   output reg               s2_awready , 
   input    [ADRS_BITS-1:0] s2_awaddr  , 
   input     [LEN_BITS-1:0] s2_awlen   , 
   input             [ 2:0] s2_awsize  , 
   input      [ID_BITS-1:0] s2_awid    , 
   input             [ 1:0] s2_awburst , 
   input             [ 3:0] s2_awcache , 
   input             [ 1:0] s2_awlock  , 
   input             [ 2:0] s2_awprot  , 
   input             [ 3:0] s2_awqos   , 
                           
   // write data           
   input                    s2_wvalid  , 
   output reg               s2_wready  , 
   input    [DATA_BITS-1:0] s2_wdata   , 
   input             [ 3:0] s2_wstrb   , 
   input                    s2_wlast   , 
   input      [ID_BITS-1:0] s2_wid     , 

   // write response    
   output                   s2_bvalid  , 
   input                    s2_bready  , 
   output           [ 1 :0] s2_bresp   , 
   output     [ID_BITS-1:0] s2_bid     , 


   //
   // read interface
   //
   
   // read address 
   input                    s2_arvalid, 
   output reg               s2_arready, 
   input   [ADRS_BITS-1: 0] s2_araddr , 
   input      [ID_BITS-1:0] s2_arid   , 
   input     [LEN_BITS-1:0] s2_arlen  , 
   input            [ 2: 0] s2_arsize , 
   input            [ 1: 0] s2_arburst, 
   input            [ 3: 0] s2_arcache, 
   input            [ 1: 0] s2_arlock , 
   input            [ 2: 0] s2_arprot , 
   input            [ 3: 0] s2_arqos  , 

   // read data
   output                   s2_rvalid , 
   input                    s2_rready , 
   output   [DATA_BITS-1:0] s2_rdata  , 
   output                   s2_rlast  , 
   output     [ID_BITS-1:0] s2_rid    , 
   output           [ 1 :0] s2_rresp  , 
 
///////////////////////////////////////////////////////////   
 
   /*****************\
    *               *
    *  MASTER PORT  *
    *               *
   \*****************/
   
   // ID is bigger here 

   //
   // write interface
   //
   
   // write address 
   
   output reg               m_awvalid , 
   input                    m_awready , 
   output reg [ADRS_BITS-1:0] m_awaddr  , 
   output reg [LEN_BITS-1:0] m_awlen   , 
   output reg        [ 2:0] m_awsize  , 
   output reg [ID_BITS+1:0] m_awid    , 
   output reg        [ 1:0] m_awburst , 
   output reg        [ 3:0] m_awcache , 
   output reg        [ 1:0] m_awlock  , 
   output reg        [ 2:0] m_awprot  , 
   output reg        [ 3:0] m_awqos   , 
                           
   // write data           
   output reg               m_wvalid  , 
   input                    m_wready  , 
   output reg [DATA_BITS-1:0] m_wdata , 
   output reg        [ 3:0] m_wstrb   , 
   output reg               m_wlast   , 
   output reg [ID_BITS+1:0] m_wid     , 

   // write response    
   input                    m_bvalid  , 
   output reg               m_bready  , 
   input            [ 1 :0] m_bresp   , 
   input      [ID_BITS+1:0] m_bid     , 


   //
   // read interface
   //
   
   // read address 
   output reg                m_arvalid, 
   input                     m_arready, 
   output reg [ADRS_BITS-1:0] m_araddr , 
   output reg  [ID_BITS+1:0] m_arid   , 
   output reg [LEN_BITS-1:0] m_arlen  , 
   output reg        [ 2: 0] m_arsize , 
   output reg        [ 1: 0] m_arburst, 
   output reg        [ 3: 0] m_arcache, 
   output reg        [ 1: 0] m_arlock , 
   output reg        [ 2: 0] m_arprot , 
   output reg        [ 3: 0] m_arqos  , 

   // read data
   input                     m_rvalid , 
   output reg                m_rready , 
   input     [DATA_BITS-1:0] m_rdata  , 
   input                     m_rlast  , 
   input       [ID_BITS+1:0] m_rid    , 
   input             [ 1 :0] m_rresp  
);

  /*****************\
   *               *
   *  WRITE PORTS  *
   *               *
  \*****************/

// 
// The data from slaves to master is controlled here
// We handle them combined. That is: address and data are
// both handled by one statemachine. 
// 
// The principle is that once a slave is selected the
// address and data are passed between slave and master.
// The address and/or data port are closed independantly
// when the transaction of that port has finished. 
// When both have finished a new slave port is selected.
// 
// Slave port slection is done on a round-robin basis. 
//   
// The write response flow and read data return flow are controlled
// by the ID bits. There is not even a statemachine for any.
// 

localparam PASS_AD = 2'b00,
           PASS_A  = 2'b01,
           PASS_D  = 2'b10;

reg  [1:0]  aw_port;   // Which port we service 0 , 1  or 2
reg  [1:0]  aw_state;  // Where we are 


wire adrs_end = m_awvalid & m_awready;
wire data_end = m_wvalid  & m_wready & m_wlast;
wire pass_adrs = (aw_state==PASS_AD) | (aw_state==PASS_A);
wire pass_data = (aw_state==PASS_AD) | (aw_state==PASS_D);

   //
   // Write port mux 
   //
   // Here we also handle extra ID bits
   always @ ( * )
   begin

      // Write address bus 
      if (aw_port==2'b00)
      begin // Address port 0 selected
         // Write address bus 
         m_awvalid = s0_awvalid & pass_adrs;
         m_awaddr  = s0_awaddr ;
         m_awlen   = s0_awlen  ;
         m_awsize  = s0_awsize ;
         m_awid    = {s0_awid,2'b00}; // Add ID bits
         m_awburst = s0_awburst;
         m_awcache = s0_awcache;
         m_awlock  = s0_awlock ;
         m_awprot  = s0_awprot ;
         m_awqos   = s0_awqos  ;        
         //m_awuser  = s0_awuser; 
         
         s0_awready = m_awready & pass_adrs;
         s1_awready = 1'b0;
         s2_awready = 1'b0;
         
      end
      else if (aw_port==2'b01)
      begin // Address port 1 selected
         m_awvalid = s1_awvalid;
         m_awaddr  = s1_awaddr ;
         m_awlen   = s1_awlen  ;
         m_awsize  = s1_awsize ;
         m_awid    = {s1_awid,2'b01}; // Add ID bits 
         m_awburst = s1_awburst;
         m_awcache = s1_awcache;
         m_awlock  = s1_awlock ;
         m_awprot  = s1_awprot ;
         m_awqos   = s1_awqos  ;
         
         s0_awready = 1'b0;
         s1_awready = m_awready & pass_adrs;
         s2_awready = 1'b0;
      end
      else
      begin // Address port 2 selected
         m_awvalid = s2_awvalid & pass_adrs;
         m_awaddr  = s2_awaddr ;
         m_awlen   = s2_awlen  ;
         m_awsize  = s2_awsize ;
         m_awid    = {s2_awid,2'b10}; // Add ID bits 
         m_awburst = s2_awburst;
         m_awcache = s2_awcache;
         m_awlock  = s2_awlock ;
         m_awprot  = s2_awprot ;
         m_awqos   = s2_awqos  ;
         
         s0_awready = 1'b0;
         s1_awready = 1'b0;
         s2_awready = m_awready & pass_adrs;
     end
      
      // Write data bus 
      if (aw_port==2'b00)
      begin // Data port 0 selected 
         m_wvalid  = s0_wvalid & pass_data;
         m_wdata   = s0_wdata;
         m_wstrb   = s0_wstrb;
         m_wlast   = s0_wlast;
         m_wid     = {s0_wid,2'b00}; // add ID bits 
         
         s0_wready = m_wready & pass_data;
         s1_wready  = 1'b0;
         s2_wready  = 1'b0;
      end
      else if (aw_port==2'b01)
      begin // Data port 1 selected        
         m_wvalid  = s1_wvalid & pass_data;
         m_wdata   = s1_wdata;
         m_wstrb   = s1_wstrb;
         m_wlast   = s1_wlast;
         m_wid     = {s1_wid,2'b01}; // add ID bits  
         
         s0_wready = 1'b0;
         s1_wready = m_wready & pass_data;
         s2_wready = 1'b0;

      end
      else 
      begin // Data port 2 selected        
         m_wvalid  = s2_wvalid  & pass_data;
         m_wdata   = s2_wdata;
         m_wstrb   = s2_wstrb;
         m_wlast   = s2_wlast;
         m_wid     = {s2_wid,2'b10}; // add ID bits  
         
         s0_wready = 1'b0;
         s1_wready = 1'b0;
         s2_wready = m_wready & pass_data;

      end
        
   end // always 

   // 
   // Check if there is a request on the 'other' (not selected port)
   // AXI specifies that we may NOT retract an active valid 
   // (and its associated values) until a ready arrives
   // Thus we are allowed to switch only if
   // - There is no valid
   // - There is a valid & ready
   //
   // Also on write data we do NOT interrupt a burst
   // Thus we can switch only if there is also a 'last'
 
  always @(posedge clk or negedge reset_n)
   begin
      if (!reset_n)
      begin
         aw_port <= 2'b00;
         aw_state <= PASS_AD;
      end
      else
      begin
 
         case (aw_state)         
         PASS_AD: 
            // address and data are passed on
            // Next depends on who finishes first
            begin
               case ( {adrs_end,data_end} )
               2'b00 : // No address/data cycles ending 
                     // if nothing pending we can switch 
                     if (!m_awvalid & !m_wvalid)
                        select_next_aw_task; 
               2'b01 : // Done with data. wait for address 
                     aw_state <= PASS_A;
               2'b10 : // Done with address. wait for data 
                     aw_state <= PASS_D;
               2'b11 : // Both done. 
                     select_next_aw_task;
               endcase
            end // case PASS_AD
               
         PASS_A: 
            if (adrs_end)
               select_next_aw_task;
            
         PASS_D:
            if (data_end)
               select_next_aw_task;
               
         endcase
      end // clocked 
   end // always 


task select_next_aw_task;
begin
   if (aw_port==2'b00)
   begin
      if (s1_awvalid)
         aw_port <= 2'b01;
      else
      if (s2_awvalid)
         aw_port <= 2'b10;
   end
   else if (aw_port==2'b01)
   begin
      if (s2_awvalid)
         aw_port <= 2'b10;
      else
      if (s0_awvalid)
         aw_port <= 2'b00;
   end
   else
   begin
      if (s0_awvalid)
         aw_port <= 2'b00;
      else
      if (s1_awvalid)
         aw_port <= 2'b01;
   end
   aw_state <= PASS_AD;
end
endtask
   
   //
   // Write response handling
   //
   
   // The return ready/valid is controlled by the ID bits
   // Other signals are returned distributed.
   // That means they are send back to all ports relying
   // on the 'valid' signal to be picked up by one only.
   // This is definitely not the best for power consumption
   // or routing load but it is the least amount of logic
   
   assign s0_bvalid = m_bvalid & (m_bid[1:0]==2'b00);
   assign s0_bresp  = m_bresp;  
   assign s0_bid    = m_bid>>2; // Drop the ID routing bits we added

   assign s1_bvalid = m_bvalid & (m_bid[1:0]==2'b01);
   assign s1_bresp  = m_bresp;  
   assign s1_bid    = m_bid>>2; // Drop the ID routing bit we added

   assign s2_bvalid = m_bvalid & (m_bid[1:0]==2'b10);
   assign s2_bresp  = m_bresp;  
   assign s2_bid    = m_bid>>2; // Drop the ID routing bits we added
   
   // Ready is also routed using the id bits
   // But the id bits may only be used if there is a valid
   always @( * )
   begin
      if (!m_bvalid)
         m_bready = 1'b0;
      else
         if (m_bid[1:0]==2'b00)
            m_bready = s0_bready;
         else  if (m_bid[1:0]==2'b01)
            m_bready = s1_bready;
         else
            m_bready = s2_bready;
   end
  
   
 
  /****************\
   *              *
   *  READ PORTS  *
   *              *
  \****************/

//
// Same as write port but we only have an address
// 
// The read response flow is controlled by the ID bits
// There is not even a statemachine for it.
// 


reg [1:0] ar_port;  // Which port we service 0, 1 or 2

   
   //
   // Read address port mux 
   //
   // Here we also handle extra ID bits
   always @ ( * )
   begin
      if (ar_port==2'b00)
      begin
         // Read address bus 0
         m_arvalid = s0_arvalid;
         m_araddr  = s0_araddr ;
         m_arlen   = s0_arlen  ;
         m_arsize  = s0_arsize ;
         m_arid   = {s0_arid,2'b00}; // Add ID bits
         m_arburst = s0_arburst;
         m_arcache = s0_arcache;
         m_arlock  = s0_arlock ;
         m_arprot  = s0_arprot ;
         m_arqos   = s0_arqos  ;
         
         s0_arready = m_arready;
         s1_arready = 1'b0;
         s2_arready = 1'b0;
      end
      else if (ar_port==2'b01)
      begin
         // Read address bus 1
         m_arvalid = s1_arvalid;
         m_araddr  = s1_araddr ;
         m_arlen   = s1_arlen  ;
         m_arsize  = s1_arsize ;
         m_arid   = {s1_arid,2'b01}; // Add ID bits
         m_arburst = s1_arburst;
         m_arcache = s1_arcache;
         m_arlock  = s1_arlock ;
         m_arprot  = s1_arprot ;
         m_arqos   = s1_arqos  ;
         
         s0_arready = 1'b0;
         s1_arready = m_arready;
         s2_arready = 1'b0;
      end
      else
      begin
         // Read address bus 2
         m_arvalid = s2_arvalid;
         m_araddr  = s2_araddr ;
         m_arlen   = s2_arlen  ;
         m_arsize  = s2_arsize ;
         m_arid   = {s2_arid,2'b10}; // Add ID bits
         m_arburst = s2_arburst;
         m_arcache = s2_arcache;
         m_arlock  = s2_arlock ;
         m_arprot  = s2_arprot ;
         m_arqos   = s2_arqos  ;
         
         s0_arready = 1'b0;
         s1_arready = 1'b0;
         s2_arready = m_arready;
       end
   end // always 
   
   //
   // See comment at write code
   //
   always @(posedge clk or negedge reset_n)
   begin
      if (!reset_n)
         ar_port <= 2'b00;
      else
      begin
      
         if (!m_arvalid | (m_arvalid & m_arready))
         begin
            if (ar_port==2'b00)
            begin
               if (s1_arvalid)
                  ar_port <= 2'b01;
               else
               if (s2_arvalid)
                  ar_port <= 2'b10;
            end
            else if (ar_port==2'b01)
            begin
               if (s2_arvalid)
                  ar_port <= 2'b10;
               else
               if (s0_arvalid)
                  ar_port <= 2'b00;
            end
            else
            begin
               if (s0_arvalid)
                  ar_port <= 2'b00;
               else
               if (s1_arvalid)
                  ar_port <= 2'b01;
            end
         end
       end // clocked
   end // always 
             
                     
   //
   // Read response handling
   //
   
   // The return ready/valid is controlled by the ID bits
   // Other signals are returned distributed.
   // That means they are send back to all ports relying
   // on the 'valid' signal to be picked up by one only.
   // This is definitely not the best for power consumption
   // or routing load but it is the least amount of logic.
   // (This also means the return stream can switch to the
   //  other port mid-streams but, with no valid, the slave ports
   //  will not notice that)
  
   assign s0_rvalid = m_rvalid & (m_rid[1:0]==2'b00);
   assign s0_rdata  = m_rdata; 
   assign s0_rlast  = m_rlast; 
   assign s0_rresp  = m_rresp;  
   assign s0_rid    = m_rid>>2; // Drop the ID routing bits we added

   assign s1_rvalid = m_rvalid &  (m_rid[1:0]==2'b01);
   assign s1_rdata  = m_rdata; 
   assign s1_rlast  = m_rlast; 
   assign s1_rresp  = m_rresp;  
   assign s1_rid    = m_rid>>2; // Drop the ID routing bits we added
 
   assign s2_rvalid = m_rvalid &  (m_rid[1:0]==2'b10);
   assign s2_rdata  = m_rdata; 
   assign s2_rlast  = m_rlast; 
   assign s2_rresp  = m_rresp;  
   assign s2_rid    = m_rid>>2; // Drop the ID routing bits we added
  
   // Ready is also routed using the id bit 
   // But the id bit may only be used if there is a valid
   always @( * )
   begin
      if (!m_rvalid)
         m_rready = 1'b0;
      else
         if (m_rid[1:0]==2'b00)
            m_rready = s0_rready;
         else  if (m_rid[1:0]==2'b01)
            m_rready = s1_rready;
         else
            m_rready = s2_rready;
   end
  
 
endmodule    
